import numpy as np
import pandas as pd
import seaborn as sns
import warnings

from scipy.io import arff
from sklearn.metrics import classification_report, confusion_matrix

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, LabelEncoder
from sklearn.tree import DecisionTreeClassifier

from utilities import plot_counts
from utilities import plot_features, ordinal_encode, normalize, plot_confusion_matrix

warnings.filterwarnings("ignore")


# Constants
y_data = 'class'
read_csv = False
csv_path = "nsl-kdd-dataset/" + "fruits_dataset.csv"
arff_path = "nsl-kdd-dataset/" + "KDDTest+.arff"
sc = StandardScaler()
enc = LabelEncoder()
model_name = "Decision Tree"

if read_csv:
    # Getting data frame (CSV)
    df = pd.read_csv(csv_path)
else:
    # Getting data frame (ARFF)
    data, meta = arff.loadarff(arff_path)
    df = pd.DataFrame(data)
    df = df.applymap(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)

X = df.select_dtypes(include=[np.number])
y = df[y_data]

is_threat = df[y_data].unique()
if len(is_threat) != 2:
    raise Exception("Target must be a binary decision.")

ordinal_encode(df= df, categories = is_threat, target = y_data)
normalize(df, y_data, sc, enc)

sns.countplot(x = y_data, data = df)
plot_counts(model_name=model_name)

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = 0.2, random_state = 42)

def train(dct_model):
    dct_model.fit(X_train, y_train)
    y_prediction = dct_model.predict(X_test)
    print("Classification report: \n", classification_report(y_test, y_prediction))
    plot_confusion_matrix(confusion_matrix = confusion_matrix(y_test, y_prediction),
                          accuracy = dct_model.score(X_test, y_test),
                          model_name=model_name)

dtc = DecisionTreeClassifier()
train(dtc)
features = pd.DataFrame(dtc.feature_importances_,
                        index= X.columns,
                        columns=['Importance']).sort_values(by='Importance', ascending=False)
plot_features(features, model_name=model_name)

# Different parameters
# Criterion modifiers ('criterion'):
# 'gini' is computationally faster because it does not involve computing logarithms.
# 'entropy' tends to work better when class distributions are highly imbalanced.
# 'log_loss' is more useful when you care about predicting class probabilities accurately.

# Cost Complexity Pruning ('ccp_alpha'):
# The ccp_alpha parameter controls the pruning process in a decision tree.
# Pruning is the process of removing parts of the tree that do not provide significant improvement in prediction,
# to reduce overfitting.
# Higher ccp_alpha -> more pruning -> simpler tree and vice versa

# Max Depth ('max_depth'):
# Limits how much a tree can grow
# Small depth -> simpler, less overfitting and vice versa

# Minimum Samples Fit ('min_samples_split'):
# Determines how the min. of samples to split a node
# Higher value -> lesser splits -> simpler tree and vice versa

# Random state ('random_state'):
# Allows reproducibility

# What is overfitting?
# Machine learning model "memorizes" the training data, rather than understanding the underlying pattern
# Different parameters

dtc2 = DecisionTreeClassifier(criterion ='entropy', ccp_alpha = 0.04)
train(dtc2)
features = pd.DataFrame(dtc2.feature_importances_,
                        index= X.columns,
                        columns=['Importance']).sort_values(by='Importance', ascending=False)
plot_features(features, model_name=model_name)

def predict(prediction_input):
    if len(prediction_input) == 0:
        return
    input_data = pd.DataFrame(prediction_input, columns = X.columns)
    return dtc.predict(input_data)
