import numpy as np
import pandas as pd
import seaborn as sns
from scipy.io import arff

import warnings
from sklearn.preprocessing import StandardScaler, LabelEncoder

from utilities import ordinal_encode, heat_map, plot_features, plot_confusion_matrix, normalize, print_high_confidence_samples, plot_counts

warnings.filterwarnings("ignore")

from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import classification_report, confusion_matrix

# Constants

X_data = 'count'
y_data = 'class'
read_csv = False
csv_path = "nsl-kdd-dataset/" + "fruits_dataset.csv"
arff_path = "nsl-kdd-dataset/" + "KDDTrain+.arff"
sc = StandardScaler()
enc = LabelEncoder()
model_name = "Logistic Regression"

if read_csv:
    # Getting data frame (CSV)
    df = pd.read_csv(csv_path)
else:
    # Getting data frame (ARFF)
    data, meta = arff.loadarff(arff_path)
    df = pd.DataFrame(data)
    df = df.applymap(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)

# Data inspection and pre-processing
print("Before: \n", df)

is_threat = df[y_data].unique()
if len(is_threat) != 2:
    raise Exception("Logistic Regression only works for binary classification.")

numerical_columns = df.select_dtypes(include=np.number).columns
label_columns = df.select_dtypes(include=object, exclude=np.number).columns

# Normalize data
ordinal_encode(df= df, categories = is_threat, target = y_data)
normalize(df, y_data, sc, enc)
print("After: \n", df)

# Correlation
heat_map(df, model_name=model_name)

# Count plot
sns.countplot(x = y_data, data = df)
plot_counts(model_name=model_name)

# Data preparation
X = df.select_dtypes(include = np.number).drop(columns = y_data) # X must be independent
y = df[[y_data]] # y must ideally be dependent on X, and in case of logistic regression, must be a binary decision

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = 0.2, random_state = 42)

# 'penalty'
# Helps preventing overfitting.
# l2 (default), l1, elasticnet (l1 + l2), none (not recommended)

# 'C'
# Variable that determines "how much is the model allowed to fit to the data?"
# smaller C -> simpler model (less overfitting) and vice versa

# 'solver'
# Algorithms that optimize logistic regression
# 'lbfgs' (default) and works well with l2
# 'liblinear' good for small datasets and works well with l1
# 'saga' handles larger datasets, supports all penalty types
# 'newton-cg', 'sag' useful for larger datasets

# 'l1-ratio'
# Support utility for 'elasticnet', determines the ratio of l1 and l2 regularization

# 'max_iter'
# Maximum of iterations for the solver, sometimes needed if the data is large and the model struggles to converge
# Ergo larger nsl-kdd-dataset or elasticnet may require a larger 'max_iter'

# 'class_weight'
# Handles class imbalance by giving unrepresented classes higher weights. To enable use 'balanced'

# 'random_state'
# Allows reproducibility

model = LogisticRegression()
model.fit(X_train, y_train)
y_prediction = model.predict(X_test)
plot_confusion_matrix(confusion_matrix = confusion_matrix(y_test, y_prediction),
                      accuracy = model.score(X_test, y_test),
                      model_name=model_name)
print(classification_report(y_test, y_prediction))

def predict(prediction_input):
    if len(prediction_input) == 0:
        return
    input_df = pd.DataFrame(prediction_input, columns = X.columns)
    input_df[numerical_columns] = sc.transform(input_df[numerical_columns])
    return ["anomaly" if x == 1 else "normal" for x in model.predict(input_df)]

print_high_confidence_samples(model, X)

features = pd.DataFrame(model.coef_[0], index= X.columns, columns=['Importance']).sort_values(by='Importance', ascending=False)

plot_features(features, "Positive coefficient = More probable to land in class 1\n"
                        "Negative coefficient = Less probable to land in class 1",
              model_name=model_name)
