import numpy as np
import pandas as pd
from scipy.io import arff

import warnings
from sklearn.preprocessing import StandardScaler, LabelEncoder

from utilities import ordinal_encode, heat_map, plot_features, plot_confusion_matrix, normalize, print_high_confidence_samples

warnings.filterwarnings("ignore")

from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import classification_report, confusion_matrix

# Constants
X_data = 'count'
y_data = 'class'
arff_path = "nsl-kdd-dataset/" + "KDDTrain+.arff"
sc = StandardScaler()
enc = LabelEncoder()
model_name = "Random Forest"

# Getting data frame (ARFF)
data, meta = arff.loadarff(arff_path)
df = pd.DataFrame(data)
df = df.applymap(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)

# Data inspection and pre-processing
print("Before: \n", df)

is_threat = df[y_data].unique()
if len(is_threat) != 2:
    raise Exception("Random Forest supports both binary and multi-class classification.")

numerical_columns = df.select_dtypes(include=np.number).columns
label_columns = df.select_dtypes(include=object, exclude=np.number).columns

# Normalize data
ordinal_encode(df= df, categories = is_threat, target = y_data)
normalize(df, y_data, sc, enc)
print("After: \n", df)

# Correlation
heat_map(df, model_name=model_name)

# Data preparation
X = df.select_dtypes(include = np.number).drop(columns = y_data)  # X must be independent
y = df[[y_data]]  # y must ideally be dependent on X

X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

# Initialize and train the Random Forest model
model = RandomForestClassifier(n_estimators=100, max_depth=None, random_state=42)
model.fit(X_train, y_train.values.ravel())

# Predictions
y_prediction = model.predict(X_test)

# Plot Confusion Matrix
plot_confusion_matrix(confusion_matrix=confusion_matrix(y_test, y_prediction),
                      accuracy=model.score(X_test, y_test),
                      model_name=model_name)

print(classification_report(y_test, y_prediction))

# Prediction function
def predict(prediction_input):
    if len(prediction_input) == 0:
        return
    input_df = pd.DataFrame(prediction_input, columns=X.columns)
    input_df[numerical_columns] = sc.transform(input_df[numerical_columns])
    return ["anomaly" if x == 1 else "normal" for x in model.predict(input_df)]

print_high_confidence_samples(model, X)

# Feature Importance Plot
features = pd.DataFrame(model.feature_importances_, index=X.columns, columns=['Importance']).sort_values(by='Importance', ascending=False)

plot_features(features, "Higher importance = More impact on classification", model_name=model_name)
